const Times = require("../public/times");
const Decimal = require("decimal");
const Init = require("./init");

/**
 * 查询规定窗口期内的总报块奖励，查询规定窗口期各个挖矿账户24小时平均算力
 * 通过各个账户24小时平均算力，计算出各个账户的算力占比
 * 根据算力占比对总报块奖励进行分配
 * 将分配的结果写入wallet_in表
 */
class Distribution extends Init {
  constructor(coin) {
    const method = "distribution";
    super(coin, method);
  }
  /**
   * 计算每个用户最终得分，满分为100分
   * @param {Array} alluser_mhs24h 每个用户过去24小时平均算力
   * @param {Number} hash_percent 24小时平均算力权重
   * @returns 每个用户最终得分
   */
  score(alluser_mhs24h, hash_percent = 1) {
    // 提取 mhs24h 数值
    const hashrateValues = alluser_mhs24h.map((obj) => obj.mhs24h);
    // 计算总和
    const totalHashrate = hashrateValues.reduce((sum, value) => sum + value, 0);
    const result = {};
    // 计算每个用户的算力占比
    for (let { user, mhs24h } of alluser_mhs24h) {
      result[user] = Number(((mhs24h / totalHashrate) * hash_percent).toFixed(4));
    }
    return result;
  }

  /**
   * 查询指定日期用户的24小时总算力
   * @param {Date} date 采用yyyy-MM-dd格式（只支持0时刻），例如"2024-11-11"
   */
  async query_last_day_hashrate(date) {
    try {
      const sql = `SELECT user, SUM(mhs24h) AS mhs24h FROM ${this.coin}_mhsv2 WHERE date = ? GROUP BY user;`;
      const data = await this.hashratedb.exec(sql, [date + " 00:00:00"]);
      if (data.length === 0) {
        return false;
      } else {
        const result = [];
        for (let item of data) {
          const { user, mhs24h } = item;
          result.push({ user, mhs24h: Number(mhs24h) });
        }
        return result;
      }
    } catch (err) {
      throw err;
    }
  }

  /**
   * 查询指定日期报块总奖励
   * @param {Date} start_time 采用yyyy-MM-dd格式，例如"2024-11-11"
   * @param {Date} end_time 同上
   * @return {Number} 昨日总奖励
   */
  async query_last_day_reward(start_time, end_time) {
    try {
      const sql = `SELECT MAX(height) AS max_height, SUM(reward) AS reward FROM ${this.coin}_blkreportprofitv2 WHERE date >= ? AND date < ? AND state = ?;`;
      const data = await this.distributiondb.exec(sql, [start_time, end_time, 1]);
      if (data.length === 0 || !data[0].reward) {
        return 0;
      } else {
        data[0].reward = Number(data[0].reward);
        return data;
      }
    } catch (err) {
      throw err;
    }
  }
  /**
   * 校验昨天所有报块是否成熟
   * @param {Date} start_time 采用yyyy-MM-dd hh:mm:ss格式，例如"2024-11-11 00:00:00"
   * @param {Date} end_time 同上
   * @returns
   */
  async query_last_day_if_mature(start_time, end_time) {
    try {
      const sql = `SELECT count(*) AS count FROM ${this.coin}_blkreportprofitv2 WHERE date >= ? AND date < ? AND state = ?;`;
      let data;

      // 使用循环替代递归
      do {
        await this.sleep(1000 * 60 * 15); // 等待 15 分钟
        data = await this.distributiondb.exec(sql, [start_time, end_time, 0]);

        // 动态获取当前小时
        const currentHour = Number(Times.times()[4]);
        if (this.coin === "rxd") {
          if (currentHour >= 9) {
            console.log("已超过凌晨 9 点，停止检查。");
            return false;
          }
        } else {
          if (currentHour >= 4) {
            console.log("已超过凌晨 4 点，停止检查。");
            return false;
          }
        }
      } while (data[0].count !== 0);

      return true;
    } catch (err) {
      throw err;
    }
  }

  /**
   * 查询当前挖矿账户状态，包括起付额和钱包地址
   * @returns
   */
  async query_users_address() {
    try {
      const sql = `
          SELECT
            a.miner_user AS 'user', 
            b.balance AS 'address',
            b.amount AS 'amount',
            b.active AS 'state'
          FROM 
            user_account_balance b
            LEFT JOIN user_miner_account a ON b.ma_id = a.id 
          WHERE 
            a.coin = ? 
            AND b.status = 0;`;
      const data = await this.users_addresses.exec(sql, [this.coin]);
      if (!data || data.length === 0) {
        return false;
      }
      return data;
    } catch (err) {
      throw err;
    }
  }

  /**
   * 校验报块
   * @param {Number} height
   * @returns
   */
  async verify_block(height) {
    try {
      const data = await this.node.verify_block(height, this.REPORT_ADDRESS);
      if (data && typeof data === "object") {
        return this.node.block(data);
      }
      return false;
    } catch (err) {
      throw err;
    }
  }

  /**
   * 将漏掉的报块插入到数据库中
   * @param {Array} data [{time: 1708256800, height: 123456, hash: "0x1234567890", reward: 100, fees: 10}]
   */
  async insert_blkreportprofit(data) {
    try {
      let sql = `INSERT INTO ${this.coin}_blkreportprofitv2 (date, height, hash, reward, fees, state) VALUES `;
      const values = [];
      data.forEach((item) => {
        const { time, height, hash, block_reward, block_fees } = item;
        sql += `(?,?,?,?,?,?), `;
        values.push(Times.utcTime(time * 1000), height, hash, block_reward, block_fees, 1);
      });
      sql = sql.slice(0, -2);
      await this.distributiondb.exec_transaction(sql, values);
    } catch (err) {
      throw err;
    }
  }

  /**
   * 校验昨日是否有漏掉报块的情况
   * @param {Date} date 当前时间
   * @returns
   */
  async check_last_data_blk(date) {
    try {
      const ts = new Date(date).valueOf() - 86400000; // 前一天
      const yMd = Times.utcTime(ts).split(" ")[0]; // "2024-10-11"
      const ymd = yMd.split("-"); // ["2024", "10", "11"]
      const table_name = `${this.coin}_pool_blkstats_${ymd[0]}${ymd[1]}${ymd[2]}`;
      const confirm_if_table = `SHOW TABLES LIKE '${table_name}';`;
      const confirm_result = await this.pooldb.exec(confirm_if_table);
      if (!confirm_result || confirm_result.length === 0) {
        console.log(`pool_blkstats表未更新，退出本次执行，请手动校验`);
        return false;
      }
      const heights = [];
      const query_blk_pool_sql = `SELECT height FROM ${table_name} WHERE DATE(date) >= ?;`;
      // const [master_data, slave_data] = await Promise.all([this.pooldb.exec(query_blk_pool_sql, [yMd]), this.pooldb_slave.exec(`SELECT height FROM ${this.coin}_pool_blkstats WHERE DATE(date) >= ?;`, [yMd])])
      // const pool_data = master_data.concat(slave_data);
      const pool_data = await this.pooldb.exec(query_blk_pool_sql, [yMd]);
      for (let item of pool_data) {
        heights.push(item.height);
      }
      const blkreport_heighs = [];
      const query_blkreport_sql = `SELECT height FROM ${this.coin}_blkreportprofitv2 WHERE DATE(date) = ? AND state = ?;`;
      const blkreport_data = await this.distributiondb.exec(query_blkreport_sql, [yMd, 1]);
      for (let item of blkreport_data) {
        blkreport_heighs.push(item.height);
      }
      const setB = new Set(blkreport_heighs);
      const need_check_heights = heights.filter((item) => !setB.has(item));

      if (need_check_heights.length === 0) {
        console.log(`${this.coin}check 完成，没有需要重新校验的区块`);
        return true;
      }
      const need_insert_data = [];
      for (let height of need_check_heights) {
        const result = await this.verify_block(height);
        if (result) {
          need_insert_data.push(result);
        }
      }
      if (need_insert_data.length === 0) {
        console.log(`${this.coin}check 完成，没有需要insert的区块`);
        return true;
      }
      await this.insert_blkreportprofit(need_insert_data);
      console.log(`${this.coin}check 完成，已将${this.coin}漏掉的报块全部插入blk表中！`);
      return true;
    } catch (err) {
      throw err;
    }
  }

  /**
   * 更新wallet_in表中不满足起付额的用户状态
   * @param {Array} min_amount [{"user": amount}]
   * @returns
   */
  async update_state(min_amount) {
    try {
      const sql = `SELECT user, SUM(amount) AS profit FROM wallet_in WHERE coin = ? AND state = ? GROUP BY user;`;
      const data = await this.distributiondb.exec(sql, [this.coin, 2]); // []
      if (!data || data.length === 0) {
        return;
      }
      const need_update_state = [];
      for (let item of data) {
        if (item.profit >= min_amount[item.user]) {
          need_update_state.push(item.user);
        }
      }
      if (need_update_state.length === 0) {
        console.log(`${this.coin}无需要更新状态的用户(2->0)`);
        return;
      }
      const values = [];
      let update_sql = `UPDATE wallet_in SET state = ? WHERE coin = ? AND user = ?;`;
      need_update_state.forEach(async (item) => {
        const { user } = item;
        update_sql += `(?, ?, ?), `;
        values.push(0, this.coin, user);
      });
      update_sql = update_sql.slice(0, -2);
      await this.distributiondb.exec_transaction(update_sql, values);
    } catch (err) {
      throw err;
    }
  }

  /**
   * 将最终分配数据插入wallet_in表中
   * @param {Array} data [{coin, user, address, create_date, should_out_date, max_height, amount, state}]
   */
  async insert_wallet_in(data) {
    try {
      let sql = `INSERT INTO wallet_in(coin, user, address, create_date, should_out_date, max_height, amount, state) VALUES `;
      const values = [];
      data.forEach((item) => {
        const { coin, user, address, create_date, should_out_date, max_height, amount, state } = item;
        values.push(coin, user, address, create_date, should_out_date, max_height, amount, state);
        sql += `(?,?,?,?,?,?,?,?), `;
      });
      sql = sql.slice(0, -2);
      await this.distributiondb.exec_transaction(sql, values);
    } catch (err) {
      throw err;
    }
  }
  async main(start_time, end_time) {
    try {
      const _if = await this.query_last_day_if_mature(start_time, end_time);
      if (!_if) {
        return;
      }
      const check_result = await this.check_last_data_blk(end_time);
      if (!check_result) {
        return;
      }
      const [last_day_mhs24h, last_day_reward, users_address] = await Promise.all([this.query_last_day_hashrate(end_time), this.query_last_day_reward(start_time, end_time), this.query_users_address()]);
      if (!last_day_mhs24h || !last_day_reward || !users_address || !check_result) {
        console.log(`查询错误`);
        console.log("last_day_mhs24h:", last_day_mhs24h);
        console.log("last_day_reward:", last_day_reward);
        console.log("users_address:", users_address);
        return;
      }
      const reward = (BigInt(last_day_reward[0].reward) * BigInt((1 - this.POOL_FEE) * 10000)) / BigInt(10000);
      const min_amount = {};
      const score_ratio = this.score(last_day_mhs24h, 1);
      const max_height = last_day_reward[0].max_height;
      let should_out_date; // 实际转账时间
      let accuracy; // user保留小数位数，100为2位，以此类推
      let count; // pool_account 保留小数位数
      if (this.coin === "nexa") {
        should_out_date = Times.utcTime(new Date(end_time).valueOf() + 1000 * 60 * 60 * 24 * 7);
        accuracy = 100;
        count = 2;
      } else if (this.coin === "rxd") {
        accuracy = 100;
        should_out_date = end_time;
        count = 2;
      } else if (this.coin === "alph") {
        accuracy = 0;
        should_out_date = end_time;
        count = 0;
      } else {
        should_out_date = end_time;
        accuracy = 100000000;
        count = 8;
      }
      let user_profit = BigInt(0)
      const result = [];
      let pool_account_address;
      for (let user in score_ratio) {
        // const profit = Math.floor(score_ratio[user] * reward * accuracy) / accuracy;
        const ratio = Math.round(score_ratio[user] * 10000);
        const profit = (reward * BigInt(ratio)) / 10000n;
        if (profit === 0) {
          continue;
        }
        user_profit += profit;
        for (let item of users_address) {
          if (item.user === "pool_account") {
            pool_account_address = item.address;
          }
          if (user === item.user) {
            min_amount[item.user] = item.min_amount; // 今天有变动的账户才会更新最小提现金额
            let state;
            if (profit >= item.amount && item.state === 0) {
              // 账号激活自动提现且当天收益满足起付额，state用0
              state = 0;
            } else if (profit < item.amount && item.state === 0) {
              // 账号激活自动提现但当天收益不满足起付额，state用2
              state = 2;
            } else if (profit >= item.amount && item.state === 1) {
              // 账号未激活自动提现但当天收益满足起付额，state用3
              state = 3;
            } else {
              // 账号未激活自动提现且当天收益不满足起付额，state用4
              state = 4;
            }
            result.push({ coin: this.coin, user, address: item.address, create_date: end_time, should_out_date, max_height, amount: profit, state });
          }
        }
      }
      const num1 = new Decimal(last_day_reward[0].reward);
      const num2 = new Decimal(user_profit);
      const pool_account_amount = num1.sub(num2).toString();
      result.push({ coin: this.coin, user: "pool_account", address: pool_account_address, create_date: end_time, should_out_date, max_height, amount: Number(Number(pool_account_amount).toFixed(count)), state: 0 });
      console.log(result);
      await this.insert_wallet_in(result);
      await this.update_state(min_amount);
      return;
    } catch (err) {
      throw err;
    }
  }
}

module.exports = Distribution;
