const Times = require("../public/times");
const Init = require("./init");

class Report extends Init {
  constructor(coin) {
    const method = "report";
    super(coin, method);
  }

  async query_mysql_last_height() {
    try {
      const sql = `SELECT MAX(height) AS max_height FROM ${this.coin}_blkreportprofitv2;`;
      const data = await this.distribution.exec(sql);
      if (!data || data.length === 0) {
        throw `${this.coin}当前无报块记录`;
      }
      return data[0].max_height;
    } catch (err) {
      throw err;
    }
  }

  async query_redis_last_height() {
    try {
      const data = await this.redis.get(`${this.coin}:last_check`);
      if (!data) {
        const result = await this.query_mysql_last_height();
        await this.redis.set(`${this.coin}:last_check`, result);
        console.log(`redis中无${this.coin} last_check数据，采用最后一个报块高度！`);
        return result;
      }
      return Number(data);
    } catch (err) {
      throw err;
    }
  }

  async query_chain_last_height() {
    try {
      const data = await this.node.getblockcount();
      return Number(data);
    } catch (err) {
      throw err;
    }
  }

  async insert_blkreportprofit(data) {
    try {
      let sql = `INSERT INTO ${this.coin}_blkreportprofitv2 (date, height, hash, reward, fees, state) VALUES `;
      const values = [];
      for (let item of data) {
        const { date, height, hash, reward, fees } = item;
        values.push(date, height, hash, reward.toString(), fees, 0);
        sql += `(?,?,?,?,?,?), `;
      }
      sql = sql.slice(0, -2);
      await this.distribution.exec_transaction(sql, values);
    } catch (err) {
      throw err;
    }
  }

  async main() {
    try {
      const [redis_height, chain_height] = await Promise.all([this.query_redis_last_height(), this.query_chain_last_height()]);
      // 区块链高度小于last_check高度，节点同步出错
      if (chain_height < redis_height) {
        console.log(`${this.coin}节点同步出错，节点高度${chain_height}，last_check${redis_height}`);
        return;
      }
      if (chain_height === redis_height) {
        console.log(`${this.coin}当前节点和last_check高度一致，无需校验`);
        return;
      }
      const suc_data = [];
      for (let i = redis_height + 1; i <= chain_height; i++) {
        const check_result = await this.node.verify_block(i, this.REPORT_ADDRESS);
        if (check_result) {
          const block = this.node.block(check_result);
          const { height, hash, time, block_reward, block_fees } = block;
          suc_data.push({ date: Times.utcTime(time * 1000), height, hash, reward: block_reward.toString(), fees: block_fees });
        }
      }
      if (suc_data.length === 0) {
        console.log(`${redis_height} - ${chain_height} 无报块`);
        await this.redis.set(`${this.coin}:last_check`, chain_height);
        return;
      }
      await this.insert_blkreportprofit(suc_data);
      await this.redis.set(`${this.coin}:last_check`, chain_height);
      return;
    } catch (err) {
      throw err;
    }
  }
}

class ReportEnx extends Report {
  constructor(coin) {
    const method = "report";
    super(coin, method);
  }

  async query_blkstats(height) {
    try {
      const yesterday = Times.utcTime(new Date().valueOf() - 24 * 60 * 60 * 1000)
        .split(" ")[0]
        .replace(/\-/g, "");
      const table1 = "enx_pool_blkstats";
      const table2 = `enx_pool_blkstats_${yesterday}`;
  
      const queryTableExistSql = `SHOW TABLES LIKE ?;`;
      const existData = await this.pooldb.exec(queryTableExistSql, [table2]);
  
      if (existData.length === 0) {
        const sql = `SELECT date, height, hash FROM ?? WHERE height > ?;`;
        return await this.pooldb.exec(sql, [table1, height]);
      } else {
        const sql = `SELECT date, height, hash FROM ?? WHERE height > ? 
                     UNION ALL 
                     SELECT date, height, hash FROM ?? WHERE height > ?;`;
        return await this.pooldb.exec(sql, [table1, height, table2, height]);
      }
    } catch (err) {
      throw err;
    }
  }
  

  async query_blkreportprofit() {
    try {
      const sql = `SELECT MAX(height) AS max_height FROM ${this.coin}_blkreportprofitv2;`;
      const result = await this.distribution.exec(sql);
      if (result.length === 0 || result[0].max_height === null) {
        return 0;
      } else {
        return result[0].max_height;
      }
    } catch (err) {
      throw err;
    }
  }

  async insertinto_blkreportprofit(data) {
    try {
      let sql = `INSERT INTO ${this.coin}_blkreportprofitv2 (date, height, hash, reward, fees, state) VALUES `;
      const values = [];
      for (let item of data) {
        const { date, height, hash, reward } = item;
        values.push(date, height, hash, reward, 0, 1);
        sql += `(?,?,?,?,?,?), `;
      }
      sql = sql.slice(0, -2);
      await this.distribution.exec_transaction(sql, values);
    } catch (err) {
      throw err;
    }
  }
  async main() {
    // const start_time = Times.utcTime(new Date().valueOf());
    const last_height = await this.query_blkreportprofit();
    const data = await this.query_blkstats(last_height);

    if (data.length === 0) {
      console.log(`${this.coin} 无报块`);
      return;
    }
    const block_data = []
    for(let item of data){
      const {date, height, hash} = item
      block_data.push({date, height, hash, reward:333})
    }

    await this.insertinto_blkreportprofit(block_data);
    return
  }
}

module.exports = {
  Report,
  ReportEnx,
}
